/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1995 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	ccxlib.c (Character Code Translator)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	951111	created
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>

#define ASCII	'A'
#define BIG5	'B'
#define GB	'G'
#define HZ	'H'
#define GRE	'g'
#define KOI8	'K'
#define CYR	'C'
#define ISO2022	'I'
#define EUCJP	'E'
#define JIS	'J'

extern char *GRE_TO_EUCJP();
extern char *CYR_TO_EUCJP();
extern char *KOI8_TO_EUCJP();
extern char *ISO2022_TO_JIS();

static cncode(str)
	unsigned char *str;
{	unsigned char *s,ch;
	int n8,n8s,nhz;

	n8 = 0;
	n8s = 0;
	nhz = 0;

	for( s = str; ch = *s; s++ ){
		if( ch == '~' )
			if( s[1] == '{' || s[1] == '}' )
				nhz++;

		if( ch & 0x80 ){
			n8++;
			n8s++;
		}else{
			if( n8 % 2 != 0 )
				return BIG5;
			n8 = 0;
		}
	}
	if( 2 <= nhz )
		return HZ;
	if( 0 < n8s )
		return GB;
	return ASCII;
}

static isISO2022(str)
     char *str;
{
  char *s;

  for (s=str; *s; s++) {
    if( *s == 033 ) {
      if( s[1] == '$' || s[1] == '(' || s[1] == ',')
	return 1;
    }
  }

  return 0;
}

ccx(src,dst,inspec,outspec,ctype)
	char *src,*dst,*inspec,*outspec,*ctype;
{	char incode,outcode;

	if( isISO2022(src) )      incode = ISO2022; else
	if( strstr(inspec,"gre")  ) incode = GRE;   else
	if( strstr(inspec,"cyr")  ) incode = CYR;   else
	if( strstr(inspec,"big5") ) incode = BIG5;  else
	if( strstr(inspec,"b5")   ) incode = BIG5;  else
	if( strstr(inspec,"gb")   ) incode = GB;    else
	if( strstr(inspec,"hz")   ) incode = HZ;    else
	if( strstr(inspec,"koi8") ) incode = KOI8;  else
	if( strstr(inspec,"jp")   ) incode = JIS;   else
	if( strstr(inspec,"jis")  ) incode = JIS;   else
	if( strstr(inspec,"euc")  ) incode = EUCJP; else
				    incode = cncode(src);

	outcode = 0;
	if( strstr(outspec,"jp")  ) outcode = JIS;   else
	if( strstr(outspec,"jis") ) outcode = JIS;   else
	if( strstr(outspec,"euc") ) outcode = EUCJP; else
	if( strstr(outspec,"gb")  ) outcode = GB;    else
	if( strstr(outspec,"big5")) outcode = BIG5;

	switch( incode ){
		default:
		case ASCII:   strcpy(dst,src);         break;
		case BIG5:    BIG5_TO_EUCJP(src,dst);  break;
		case GB:      GB_TO_EUCJP(src,dst);    break;
		case HZ:      HZ_TO_EUCJP(src,dst);    break;
		case GRE:     GRE_TO_EUCJP(src,dst);   break;
		case CYR:     CYR_TO_EUCJP(src,dst);   break;
		case KOI8:    KOI8_TO_EUCJP(src,dst);  break;
		case ISO2022: ISO2022_TO_JIS(src,dst); break;
	}
}

any_to_jis_file(chsfp,jisfp,inspec)
	FILE *chsfp,*jisfp,*inspec;
{	int nl;
	char chs[0x200],jis[0x200];

	for(nl = 1;;nl++){
		if( fgets(chs,sizeof(chs),chsfp) == NULL )
			break;
		ccx(chs,jis,"*","jis");
		fputs(jis,jisfp);
	}
}
