/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1995 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	ccxserv.c (Character Code Translator)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	951111	created
	960221	renamed from ccxlib.c
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>

#define BTOEJ_SIZE	(94*157*2)
#define MAX_PY_TBL	2066
#define MAX_PY_LEN	10

typedef unsigned char Mnemonic[MAX_PY_LEN];
unsigned short *GB2312JX0208[94];
unsigned short *GB2312JX0212[94];
unsigned short *GB2312UZPJ[94];
unsigned char *BtoEJ;
Mnemonic *py_tbl;
Mnemonic *py_tbl3;
static int pynelem = MAX_PY_TBL;
static int pyesize = MAX_PY_LEN;

#define TBLSIZE()  (3*(94*94*sizeof(short)) + 2*(pynelem*pyesize) + BTOEJ_SIZE)
#define WITH_CN2JP (BtoEJ != NULL)
#define NAME_CN2JP "CN-JP"

static dump_codemap(out,gb)
	FILE *out;
	unsigned short *gb[];
{	unsigned short *sp;
	int rx,cx,ch;

	for( rx = 0; rx < 94; rx++ ){
		sp = gb[rx];
		for( cx = 0; cx < 94; cx++ ){
			ch = sp[cx];
			putc(0xFF&(ch>>8),out);
			putc(0xFF& ch,    out);
		}
	}
}
static load_codemap(in,gb)
	FILE *in;
	unsigned short *gb[];
{	int rx,cx,ch;
	unsigned short *sp;

	for( rx = 0; rx < 94; rx++ ){
		sp = gb[rx] = (unsigned short*)malloc(94*sizeof(unsigned short));
		for( cx = 0; cx < 94; cx++ ){
			ch = getc(in);
			ch = (ch << 8) | getc(in);
			sp[cx] = ch;
		}
	}
}
static dump_mnemonic(out,py)
	FILE *out;
	unsigned char py[][MAX_PY_LEN];
{	int rx,cx,ch;
	unsigned char *sp;

	for( rx = 0; rx < MAX_PY_TBL; rx++ ){
		sp = py[rx];
		for( cx = 0; cx < MAX_PY_LEN; cx++ )
			putc(sp[cx],out);
	}
}
static Mnemonic *load_mnemonic(in,nelem,esize)
	FILE *in;
{	Mnemonic *mnemonic;
	int rx,sx;
	unsigned char *sp;

	mnemonic = (Mnemonic *)malloc(nelem*esize);
	for( rx = 0; rx < nelem; rx++ ){
		sp = mnemonic[rx];
		for( sx = 0; sx < esize; sx++ )
			sp[sx] = getc(in);
	}
	return mnemonic;
}
CN2JP_put(out)
	FILE *out;
{
	if( !WITH_CN2JP ){
		fprintf(stderr,"empty %s\n",NAME_CN2JP);
		return 0;
	}
	dump_codemap(out,GB2312JX0208);
	dump_codemap(out,GB2312JX0212);
	dump_codemap(out,GB2312UZPJ);
	dump_mnemonic(out,py_tbl);
	dump_mnemonic(out,py_tbl3);
	fwrite(BtoEJ,1,BTOEJ_SIZE,out);
	return TBLSIZE();
}
CN2JP_get(in)
	FILE *in;
{
	load_codemap(in,GB2312JX0208);
	load_codemap(in,GB2312JX0212);
	load_codemap(in,GB2312UZPJ);
	py_tbl = load_mnemonic(in,pynelem,pyesize);
	py_tbl3 = load_mnemonic(in,pynelem,pyesize);
	BtoEJ = (unsigned char*)malloc(BTOEJ_SIZE);
	fread(BtoEJ,1,BTOEJ_SIZE,in);
	return TBLSIZE();
}

CCX_putTable(name,out)
	char *name;
	FILE *out;
{	char header[1024],*hp;
	int leng;

	if( strcmp(name,"*") == 0 )
		return CCX_putTable(NAME_CN2JP,out);

	if( strcasecmp(name,NAME_CN2JP) != 0 )
		return 0;

	hp = header;
	sprintf(hp,"CCX-Name: %s\r\n",name);
	hp += strlen(hp);

	if( strcasecmp(name,NAME_CN2JP) == 0 ){
		sprintf(hp,"Py-size: %d * %d\r\n",MAX_PY_TBL,MAX_PY_LEN);
		hp += strlen(hp);
	}

	sprintf(hp,"\f\r\n\r\n");
	leng = strlen(header); 
	fputs(header,out);

	if( strcasecmp(name,NAME_CN2JP) == 0 )
		return leng + CN2JP_put(out);

	return leng;
}
CCX_getTable(name,in)
	char *name;
	FILE *in;
{	char line[1024],fname[1024],fvalue[1024];
	char ccxname[128];
	int leng;

	leng = 0;
	ccxname[0] = 0;

	for(;;){
		if( fgets(line,sizeof(line),in) == NULL )
			break;
		leng += strlen(line);
		if( line[0] == '\r' || line[0] == '\n' )
			break;

		fname[0] = fvalue[0] = 0;
		sscanf(line,"%[^ :]%*[: ]%[^\r\n]",fname,fvalue);

		if( strcasecmp(fname,"CCX-Name") == 0 )
			strcpy(ccxname,fvalue);
		else
		if( strcasecmp(fname,"Py-size") == 0 )
			sscanf(fvalue,"%d * %d",&pynelem,&pyesize);
	}
	if( strcasecmp(ccxname,NAME_CN2JP) == 0 )
		return leng + CN2JP_get(in);

	return leng;
}

CCX_lsTable(buff,size)
	char *buff;
{	int ntable;
	char *bp;

	ntable = 0;
	bp = buff;
	if( WITH_CN2JP ){
		sprintf(bp,"%-10s %6dbytes (%s)\n",
			NAME_CN2JP,TBLSIZE(),"Chinese {Big5,GB,HZ} to JIS");
		bp += strlen(bp);
		ntable++;
	}
	return ntable;
}
CCX_withTable (table)
	char *table;
{
	if( strcasecmp(table,NAME_CN2JP) == 0 )
		return WITH_CN2JP;
	return 0;
}
