/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1997 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	rescache.c
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	970430	created
//////////////////////////////////////////////////////////////////////#*/

#include <stdio.h>
#include <string.h>
#include "dns.h"

char *RES_HC_DIR = "/tmp/resolvy";
int   RES_HC_EXPIRE = 600;
static int cache_cantopen;

static FILE *hosts_cache(byname,nameaddr,mode,cpath)
	char *nameaddr,*mode,*cpath;
{	unsigned int hidx;
	char cdir[1024];
	FILE *fp;

	if( cache_cantopen && time(0) < cache_cantopen+RES_HC_EXPIRE )
		return NULL;
	cache_cantopen = 0;

	hidx = FQDN_hash(nameaddr);
	sprintf(cdir,"%s/%s",RES_HC_DIR,byname?"byname":"byaddr");
	sprintf(cpath,"%s/%02x",cdir,hidx%32);
	if( fp = fopen(cpath,mode) )
		return fp;

	if( *mode != 'r' ){
		mkdir(RES_HC_DIR,0755);
		mkdir(cdir,0755);
		if( (fp = fopen(cpath,mode)) == NULL )
			cache_cantopen = time(0);
	}

	return fp;
}

gethostbynameaddr_cache(dir,name,rv,rb,byname,cname)
	char *dir,*name;
	char *rv[],*rb,*cname;
{	char cpath[1024];
	FILE *cache,*ncache;
	char line[1024],*lp;
	int ac;
	int now,ctime,lines,expired;

	if( dir[0] == '/' && strcmp(RES_HC_DIR,dir) != 0 )
		RES_HC_DIR = strdup(dir);

	cache = hosts_cache(byname,name,"r",cpath);
	if( cache == NULL )
		return 0;

	now = time(0);
	expired = 0;
	ac = 0;
	for( lines = 0; fgets(line,sizeof(line),cache) != NULL; lines++ ){
		ctime = atoi(line);
		if( RES_HC_EXPIRE < now - ctime ){
			expired++;
			continue;
		}
		for( lp = line; *lp; lp++ ){
			if( *lp == ' ' ){
				lp++;
				break;
			}
		}
		ac += RES_matchLine("cache",byname,name,lp,&rv[ac],rb,cname);
	}
	fclose(cache);

	if( ac == 0 )
	if( lines/2 < expired )
	if( cache = hosts_cache(byname,name,"r",cpath) ){
		char newpath[1024];

		sprintf(newpath,"%s-%d",cpath,getpid());
		if( ncache = fopen(newpath,"w") ){
			fseek(cache,0,0);
			while( fgets(line,sizeof(line),cache) != NULL ){
				ctime = atoi(line);
				if( now - ctime <= RES_HC_EXPIRE )
					break;
			}
			while( fgets(line,sizeof(line),cache) != NULL )
				fputs(line,ncache);
			fclose(ncache);
			rename(newpath,cpath);
		}
	}
	return ac;
}

puthost_cache(nameaddr,rv,byname,cname)
	char *nameaddr,*rv[],*cname;
{	int ai;
	char *n1;
	unsigned char *a1;
	char cpath[1024];
	FILE *cache;
	int now;

	cache = hosts_cache(byname,nameaddr,"a",cpath);
	if( cache == NULL )
		return;

	now = time(0);
	if( byname ){
	    if( rv[0] == NULL ){
		fprintf(cache,"%d ",now);
		a1 = (unsigned char *)UNKNOWN_HOSTADDR;
		fprintf(cache,"%d.%d.%d.%d\t",a1[0],a1[1],a1[2],a1[3]);
		fprintf(cache,"%s\n",nameaddr);
		fflush(cache);
	    }else
	    for( ai = 0; a1 = (unsigned char *)rv[ai]; ai++ ){
		fprintf(cache,"%d ",now);
		fprintf(cache,"%d.%d.%d.%d\t",a1[0],a1[1],a1[2],a1[3]);
		if( cname && cname[0] && strcmp(cname,nameaddr) != 0 )
			fprintf(cache,"%s ",cname);
		fprintf(cache,"%s\n",nameaddr);
		fflush(cache);
	    }
	}else{
	    if( rv[0] == NULL ){
		fprintf(cache,"%d ",now);
		fprintf(cache,"%s\t%s\n",nameaddr,UNKNOWN_HOSTNAME);
		fflush(cache);
	    }else
	    for( ai = 0; n1 = rv[ai]; ai++ ){
		fprintf(cache,"%d ",now);
		fprintf(cache,"%s\t%s\n",nameaddr,n1);
		fflush(cache);
	    }
	}
	fflush(cache);
}

rem_unknown(rv,unknown,leng)
	char *rv[],*unknown;
{	int ai,ac;
	char *n1;
	char last[512];

	ac = 0;
	last[0] = 0;
	for( ai = 0; n1 = rv[ai]; ai++ ){
		if( 0 < leng ){
			if( strncmp(n1,unknown,leng) == 0 )
				continue;
		}else{
			if( strcmp(n1,unknown) == 0 )
				continue;
		}
		rv[ac++] = n1;
	}
	rv[ac] = 0;
	return ac;
}
