/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1994 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	access.c (Access Control)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	940303	created
//////////////////////////////////////////////////////////////////////#*/
#include "hostlist.h"
#include "delegate.h"
#include <ctype.h>

extern char *strdup();
extern char *getClientUserC();
#define Ident()	getClientUserC(Conn)

/*
 *	Reliable hosts for services with limited access.
 *	Normally this will be checked if the DeleGate server is on a FIREWALL.
 */
/* RELIABLE is obsoleted by PERMIT */

static HostList RelHostList   = { "RELIABLE",  16 };
static HostList ReachableList = { "REACHABLE", 16 };
static HostList IdentHostList = { "USERIDENT", 16 };

extern addHostList1();

static char authority[1024];
scan_AUTH(Conn,auth)
	char *auth;
{
	if( authority[0] != 0 )
		strcat(authority,",");
	strcat(authority,auth);
	/*fprintf(stderr,">>>>>>> %s [%s]\n",authority,auth);*/
}
static auth_match(cauth,tauth)
	char *cauth,*tauth;
{
	return strcmp(cauth,tauth) == 0;
}
static findAuth(what,proto,auth)
	char *what,*proto,*auth;
{	char xauth[1024];

	if( strcasecmp(proto,"smtp-vrfy") != 0 )
		proto = "*";

	sprintf(xauth,"%s:%s:%s",what,proto,auth);
	if( scan_List(authority,',',0,auth_match,xauth) )
		return 1;

	sprintf(xauth,"%s:%s:*",what,proto);
	if( scan_List(authority,',',0,auth_match,xauth) )
		return 1;

	return 0;
}
auth_manager(what,proto,host,user)
	char *what,*proto,*host,*user;
{	char userhost[256];

	sprintf(userhost,"%s@%s",user,host);
	return findAuth("manager",proto,userhost);
}
auth_anonftp(proto,user,pass)
	char *proto,*user,*pass;
{	char auth[256];

	if( strstr(authority,"anonftp:") == 0 )
		return 1;

	if( is_anonymous(user) || user[0] == 0 )
		return 0;
	if( strchr(user,'@') == 0 )
		return 0;
	sprintf(auth,"%s",pass);
	return findAuth("anonftp",proto,auth);
}
auth_proxy_auth()
{
	if( strstr(authority,"proxy:auth") )
		return 1;
	else	return 0;
}
auth_proxy_pauth()
{
	if( strstr(authority,"proxy:pauth") )
		return 1;
	else	return 0;
}

extern char *NOTIFY_PLATFORM;
static nplat;
NotifyPlatform(Conn,isreq)
	Connection *Conn;
{	char *host,hostb[256],*user;
	int port,match;

	if( nplat == 0 )
		nplat = makePathList("NotifyPlatform",NOTIFY_PLATFORM);

	if( isreq ){
		host = DST_HOST;
		port = DST_PORT;
		user = "-";
	}else{
		host = hostb;
		port = getClientHostPort(Conn,hostb);
		user = "-";
	}
	match = matchPath1(nplat,user,host,port);
	return match;
}
makeVia(Conn,received)
	Connection *Conn;
	char *received;
{
	if( strstr(authority,"forward") )
		ClientIF_HP(Conn,received);
	else	strcpy(received,"-");
}
makeForwarded(Conn,forwarded)
	Connection *Conn;
	char *forwarded;
{	char myuri[256],client[256];
	char myhp[256];

	if( strstr(authority,"forward") ){
		ClientIF_HP(Conn,myhp);
		sprintf(myuri,"http://%s/",myhp);
		getClientHostPort(Conn,client);
	}else{
		return 0;
		/*
		strcpy(myuri,"-");
		strcpy(client,"-");
		*/
	}
	sprintf(forwarded,"by %s (DeleGate/%s) for %s",
		myuri,(char*)DELEGATE_ver(),client);

	return 1;
}

makeAuthorization(Conn,genauth)
	Connection *Conn;
	char *genauth;
{	char *fmt,atype[128],afmt[128];
	char gauth[256],eauth[256],host[128];
	int port;
	char *dp;

	if( (fmt = strstr(authority,"authgen:")) == NULL )
		return 0;
	fmt += strlen("authgen:");
	atype[0] = afmt[0] = 0;
	sscanf(fmt,"%[^:,]:%[^,]",atype,afmt);
	atype[0] = toupper(atype[0]);

	strfConn(Conn,afmt,gauth);
	if( strchr(gauth,':') == NULL )
		strcat(gauth,":");
	gethostname(host,sizeof(host));
	strcat(gauth,"/");
	strcat(gauth,host);

	str_to64(gauth,strlen(gauth),eauth,512,1);
	if( dp = strpbrk(eauth,"\r\n") )
		*dp = 0;

	sprintf(genauth,"%s %s",atype,eauth);
	return 1;
}

makeFrom(Conn,genfrom)
	Connection *Conn;
	char *genfrom;
{	char *fmt,gfmt[128];

	genfrom[0] = 0;
	if( (fmt = strstr(authority,"fromgen")) == NULL )
		return 0;
	fmt += strlen("fromgen");

	strcpy(gfmt,"%u@%h");
	sscanf(fmt+1,":%[^,]",gfmt);
	strfConn(Conn,gfmt,genfrom);
	return 1;
}

makeClientLog(Conn,clientlog)
	Connection *Conn;
	char *clientlog;
{	char host[256],iuser[256],auser[256];
	char hfmt[128],ifmt[128],afmt[128];
	char *fmt;

	strcpy(hfmt,"%h");
	strcpy(ifmt,"%u");
	strcpy(afmt,"%U");
	if( fmt = strstr(authority,"log:") ){
		fmt += strlen("log:");
		sscanf(fmt,"%[^:,]:%[^:,]:%[^,]",hfmt,ifmt,afmt);
	}

	strfConn(Conn,hfmt,host);  if(host[0] ==0) strcpy(host,"-");
	strfConn(Conn,ifmt,iuser); if(iuser[0]==0) strcpy(iuser,"-");
	strfConn(Conn,afmt,auser); if(auser[0]==0) strcpy(auser,"-");

	sprintf(clientlog,"%s %s %s",host,iuser,auser);
}

/*
 *
 */
scan_RELIABLE(Conn,rels)
	Connection *Conn;
	char *rels;
{
	scan_commaList(rels,1,addHostList1,&RelHostList);
}
scan_REACHABLE(Conn,list)
	Connection *Conn;
	char *list;
{
	scan_commaList(list,1,addHostList1,&ReachableList);
}

/*
 *	If no RELIABLE host is specified explicitly, then suppose that any
 *	hosts is relaible.  Typical case is that the DeleGate is running
 *	on the host which belongs to a secure network within a firewall.
 */
static RELIABLE_HOST(Conn,hostlist,srchost,srcport)
	Connection *Conn;
	HostList *hostlist;
	char *srchost;
{	char *suser;

	if( hostlist->hl_cnt == 0 )
		return 1;
/*
{
char ohost[128];
getOriginator(ohost);
printf("ROUTE: %s [%s]\n",get_PATH(),ohost);
}
*/
	if( Conn->cl_auth_user[0] ){
		srchost = Conn->cl_auth_host;
		srcport = 0;
		suser = Conn->cl_auth_user;
	}else{
		suser = Ident();
/*
		if( suser == NULL ){
			char qhost[128],quser[128];
			if( get_equiv_user(srchost,srcport,qhost,quser) ){
				suser = quser;
				srchost = qhost;
				srcport = 0;
			}
		}
*/
	}
	return hostIsinList_byname(hostlist,DFLT_PROTO,srchost,srcport,suser);
}
static REACHABLE_HOST(Conn,hostlist,proto,hostname,dstport)
	Connection *Conn;
	HostList *hostlist;
	char *proto;
	char *hostname;
{
	if( hostlist->hl_cnt == 0 )
		return 1;

	return hostIsinList_byname(hostlist,proto,hostname,dstport,NULL);
}

/*
 *	Not implemented yet
 *	(This is done in RELIABLE_HOST() ?)
 */
static PERMITTED_USER(Conn,hostlist,hostname,lport,stdport)
	Connection *Conn;
	HostList *hostlist;
	char *hostname;
{
	if( hostlist->hl_cnt == 0 )
		return 1;
	return 1;
}

static PERMITTED_PORT(Conn,proto,host,port)
	Connection *Conn;
	char *proto,*host;
{	int stdport;

	if( streq(proto,"telnet") ){
	    stdport = serviceport("telnet");

if( getenv("ANYPORT") == 0 )
	    if( port != 0 && port != stdport ){
		if( ImProxy )
			sv1log("Proxy telnet follows PERMIT parameter\n");
		else
		if( streq(iSERVER_PROTO,"telnet") && port==iSERVER_PORT )
			sv1log("TELNET to non-standard port: %d\n",DFLT_PORT);
		else{
			Verbose("cannot TELNET to non-standard port %d\n",
				DFLT_PORT);
			return 0;
		}
	    }
	}
	if( port == 19 ){
		if( streq(proto,"http") || streq(proto,"gopher") ){
			sv1log("HTTP,Gopher to 19(chargen) is inhibited.\n");
			return 0;
		}
	}
	return 1;
}
static PERMITTED_PAIR(Conn,proto,dhost,dport,shost,sport)
	Connection *Conn;
	char *proto,*dhost,*shost;
{
	return DELEGATE_permit(Conn,proto,dhost,dport,shost,sport);
}

PERMITTED_ACCESS(Conn,shost,sport,stdport)
	Connection *Conn;
	char *shost;
{	int permitted = 0;

	if(RELIABLE_HOST(Conn,&RelHostList,shost,sport))
	if(PERMITTED_USER(Conn,&RelHostList,shost,sport,stdport))
	if(REACHABLE_HOST(Conn,&ReachableList,DST_PROTO,DST_HOST,DST_PORT))
	if(PERMITTED_PORT(Conn,DST_PROTO,DST_HOST,DST_PORT))
	if(PERMITTED_PAIR(Conn,DST_PROTO,DST_HOST,DST_PORT,shost,sport))
		permitted = 1;

if(!permitted)
if(LOG_GENERIC){
char *user;
if( (user = getClientUserC(Conn)) == NULL )
	user = "-";
fputLog(Conn,"Reject","%s@%s:%d; to=%s://%s:%d\n",
user,shost,sport,
DST_PROTO,DST_HOST,DST_PORT);
}

	return permitted;
}

char *scanIdent(iresp,sys,user)
	char *iresp,*sys,*user;
{	char buf[1024],*dp;

	strcpy(buf,iresp);
	if( dp = strpbrk(buf,"\r\n") )
		*dp = 0;
	user[0] = 0;
	sscanf(buf,"%*d%*[ ,]%*d%*[ :]USERID%*[ :]%[^ :]%*[ :]%s\n",sys,user);

	if( user[0] ){
		Verbose("-- ident: %s : %s\n",sys,user);
		return user;
	}else{
		Verbose("-- ident: %s",iresp);
		return 0;
	}
}

getClientHostPortAddr(Conn,rhost,raddr)
	Connection *Conn;
	char *rhost,*raddr;
{	char addrhostport[256],addr[256],host[256];
	int port;

	if( _ClientHost == NULL ){
		if( getpeerName(Conn->cl_sock,addrhostport,PN_ADDRHOSTPORT) )
		if( sscanf(addrhostport,"%s %s %d",addr,host,&port) == 3 ){
			_ClientAddr = strdup(addr);
			_ClientHost = strdup(host);
			_ClientPort = port;
		}
	}
	if( _ClientHost == NULL )
		return 0;

	if( rhost != NULL ) strcpy(rhost,_ClientHost);
	if( raddr != NULL ) strcpy(raddr,_ClientAddr);
	return _ClientPort;
}
getClientHostPort(Conn,rhost)
	Connection *Conn;
	char *rhost;
{
	return getClientHostPortAddr(Conn,rhost,NULL);
}

char *getClientUser0(Conn,hostname,porti,userbuf)
	Connection *Conn;
	char *hostname;
	char *userbuf;
{	int idsv;
	FILE *ti,*fi;
	char resp[1024],sys[128],*user;
	char *myhost;
	int myport;

	Verbose("-- ident: %s\n",hostname);
	myhost = Conn->cl_sockHOSTaddr;
	idsv = Socket1("IDENT",-1,NULL,NULL,NULL, myhost,0,hostname,113, 0,NULL,0);
	Verbose("-- ident: %d\n",idsv);

	user = NULL;
	if( 0 <= idsv ){
		ti = fdopen(idsv,"w");
		fi = fdopen(idsv,"r");
		sockHostport(Conn->cl_sock,&myport);
		fprintf(ti,"%d, %d\r\n",porti,myport);
		Verbose("-- ident: %d, %d\n",porti,myport);
		fflush(ti);
		if( fgetsTIMEOUT(resp,sizeof(resp),fi) == NULL )
			Verbose("-- ident: EOF from the server.\n");
		else{
			if( scanIdent(resp,sys,userbuf) )
				user = userbuf;
		}
		fclose(ti);
		fclose(fi);
	}
	return user;
}
static char *getPortUser(Conn,hostport,userbuf)
	Connection *Conn;
	char *hostport,*userbuf;
{	char host[1024];
	int port;

	if( sscanf(hostport,"%[^:]:%d",host,&port) == 2 )
		return getClientUser0(Conn,host,port,userbuf);
	return NULL;
}

enableClientIdent(host)
	char *host;
{
	Verbose("-- ident: ENABLE{%s}\n",host);
	scan_commaList(host,1,addHostList1,&IdentHostList);
}

#define DONT_TRY	(char*)-1
#define CANT_CONN	(char*)-2

char *getClientUserH0(Conn,hostname,hosti,porti,force)
	Connection *Conn;
	char *hostname;
{	char user[128];
	int isin;

	if( _ClientUser != NULL )
		goto EXIT;

	if( force )
		isin = 1;
	else
	if( hosti )
		isin = hostIsinList(&IdentHostList,ANYP,hostname,hosti,0,NULL);
	else	isin = hostIsinList_byname(&IdentHostList,ANYP,hostname,0,NULL);

	if( !isin )
		_ClientUser = DONT_TRY;
	else
	if( getClientUser0(Conn,hostname,porti,user) )
		_ClientUser = strdup(user);
	else	_ClientUser = CANT_CONN;

EXIT:
	if( _ClientUser == DONT_TRY || _ClientUser == CANT_CONN )
		return NULL;
	else	return _ClientUser;
}
char *getClientUserH(Conn,hostname,hosti,porti)
	Connection *Conn;
	char *hostname;
{
	return getClientUserH0(Conn,hostname,hosti,porti,0);
}

char *getClientUserC(Conn)
	Connection *Conn;
{	char host[256];
	int port;

	if( port = getClientHostPort(Conn,host) )
		return getClientUserH(Conn,host,0,port);
	return NULL;
}
char *getClientUserMbox(Conn,mbox)
	Connection *Conn;
	char *mbox;
{	char host[256],user[256];
	int port;

	strcpy(mbox,"?");
	if( port = getClientHostPort(Conn,host) ){
		if( getClientUser0(Conn,host,port,user) ){
			getFQDN(host,host);
			sprintf(mbox,"%s@%s",user,host);
			return mbox;
		}
	}
	return NULL;
}


/*
 *	PERMIT about RELAY function
 *	control accessibility of relay function
 */
scan_RELAY(Conn,relay)
	Connection *Conn;
	char *relay;
{	char realm[32],triple[1024],cmap[1024];

	realm[0] = triple[0] = 0;
	sscanf(relay,"%[^:]:%s",realm,triple);
	sprintf(cmap,"%s:relay:%s",realm,triple);
	scan_CMAP(Conn,cmap);
}
do_RELAY(Conn,what)
	Connection *Conn;
{	char realm[128];

	/* this function should be unified with service_permitted() ... */
	if( Conn->from_myself )
		return 1;

	if( 0 <= find_CMAP(Conn,"relay",realm) ){
		if( strstr(realm,"proxy") && (what & RELAY_PROXY) )
			return 1;
		if( strstr(realm,"delegate") && (what & RELAY_DELEGATE) )
			return 1;
	}
	return 0;
}

Connection *FORCE_REWRITE = (Connection*)-1;
/*
it's too heavy to be checked for each embedded URL in HTML...

isRELAYABLE(Conn,proto,hostport)
	Connection *Conn;
	char *proto,*hostport;
{
	if( Conn == FORCE_REWRITE )
		return 1;
	return do_RELAY(Conn,RELAY_DELEGATE);
}
*/

isREACHABLE(proto,hostport)
	char *proto,*hostport;
{	char host[128];
	int port;
	int svf,yes;

	port = scan_hostport(proto,hostport,host);
	svf = RES_CACHEONLY(1);
	yes = REACHABLE_HOST(NULL,&ReachableList,ANYP,host,port);
	RES_CACHEONLY(svf);
	return yes;
}

addRejectList(Conn,what,dpath,referer,auser,apass,reason)
	Connection *Conn;
	char *what,*dpath,*referer,*auser,*apass,*reason;
{	char src_host[256],*src_user;
	int src_port;

	src_port = getClientHostPort(Conn,src_host);
	src_user = getClientUserH0(Conn,src_host,0,src_port,1);
	if( src_user == NULL )
		src_user = "-";

	if( *apass != 0 )
	if( !strcaseeq(auser,"anonymous") )
	if( !strcaseeq(auser,"ftp") || !strcaseeq(DST_PROTO,"ftp") )
		apass = "*";

	if( *referer == 0 )
		referer = "-";

	putRejectList(what,
		DST_PROTO, DST_HOST,DST_PORT, dpath, referer,
		DFLT_PROTO,src_host,src_port, src_user,
		auser,apass,reason);
}
