/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1997 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	smtpgw.c (SMTP to NNTP gateway)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:

    WHAT IS THIS ?

	SMTP->{NNTP,SMTP} relay

		will be extended to

		NNTP->{NNTP,SMTP}
		SMTP->FTP
		...

    CONFIGURATION FILES

	DELEGATE_SMTPGATE/{admin,users}/To/conf  -- configuration file
	DELEGATE_SMTPGATE/{admin,users}/To/count -- counter file
	DELEGATE_SMTPGATE/{admin,users}/To/log   -- log file ?
	DELEGATE_SMTPGATE/admin/+default/conf    -- default configuration file

    QUESTION ?

	SERVER: nntp://host:port/newsgroup
		- partial inheritance becomes difficult in this format...

OWNER ... execute in the list owners effective UID ?

History:
	970829	created
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>
#include <ctype.h>
#include "delegate.h"
#define LNSIZE  1024
#define lfprintf	SMTP_lfprintf

extern char *DELEGATE_MANAGER;
extern char *strip_spaces();
extern char *strcasestr();
extern char *nextField();
extern char *malloc();
extern char *calloc();
extern FILE *SMTP_getEXPN();

#define POSTNEWS	"postnews"
#define SENDMAIL	"sendmail"
#define FTPMAIL		"ftpmail"
#define HTTPMAIL	"httpmail"

#define DEFAULT		"@default"

#define V_SENDER	"${sender}"
#define V_RECNAME	"${recipient.name}"
#define V_SUBJECT	"${subject}"
#define V_CLEANSUBJ	"${subject:hc}"
#define V_SEQNO		"${seqno}"
#define V_ERRORSTAT	"${error.status}"


#define GWTYPE			 0
#define INHERIT			 1

#define ACC_SENDER		 2
#define ACC_RECIPIENT		 3
#define ACC_FROM		 4
#define ACC_TO			 5

#define OUT_NEWSGROUPS		 6
#define OUT_DISTRIBUTION	 7
#define OUT_REPLY_TO		 8
#define OUT_TO			 9
#define OUT_SUBJECT		10
#define OUT_HEADER		11
#define OUT_FILTER		12

#define BCC			13
#define OPTION			14
#define SERVPROTO		15
#define SERVHOST		16
#define SERVPORT		17
#define COUNTER			18

#define NPARAMS			19

typedef struct {
	char	*p_abbr;
	char	 p_pfix_opt;
	char	*p_pfix;
	char	*p_name;
} PSPEC;

static PSPEC param_spec[NPARAMS] = {
	{"gw",	1,"CONTROL",	"GATEWAY-TYPE"	},
	{"gw",	1,"CONTROL",	"INHERIT"	},

	{"se",	1,"ACCEPT",	"Sender"	},
	{"re",	1,"ACCEPT",	"Recipient"	},
	{"fr",	1,"ACCEPT",	"From"		},
	{"to",	0,"ACCEPT",	"To"		},

	{"ng",	1,"OUTPUT",	"Newsgroups"	},
	{"db",	1,"OUTPUT",	"Distribution"	},
	{"rt",	1,"OUTPUT",	"Reply-To"	},
	{"ot",  1,"OUTPUT",	"To"		},
	{"os",  1,"OUTPUT",	"Subject"	},
	{"oh",  1,"OUTPUT",	"Header"	},
	{"of",  1,"OUTPUT",	"FILTER"	},

	{"bc",	1,"CONTROL",	"BCC"		},
	{"op",	1,"CONTROL",	"OPTION"	},
	{"sv",	1,"CONTROL",	"SERVER-PROTO"	},
	{"sh",	1,"CONTROL",	"SERVER-HOST"	},
	{"sp",	1,"CONTROL",	"SERVER-PORT"	},

	{"ct",	1,"CONTROL",	"COUNTER"	},
};
static param_idx(pname)
	char *pname;
{	int px;
	PSPEC *ps;
	char *dp;
	char pfix[128],name[128];

	pfix[0] = name[0] = 0;
	if( dp = strchr(pname,'/') )
		sscanf(pname,"%[^/]/%s",pfix,name);
	else	strcpy(name,pname);

	for( px = 0; px < NPARAMS; px++ ){
		ps = &param_spec[px];
		if( pfix[0] == 0 ){
			if( !ps->p_pfix_opt )
				continue;
		}else{
			if( !strcaseeq(pfix,ps->p_pfix) )
				continue;
		}
		if( strcasecmp(name,ps->p_name) == 0 )
			return px;
		if( strcasecmp(name,ps->p_abbr) == 0 )
			return px;
	}
	return -1;
}

typedef struct gateway {
	char	*g_pv[NPARAMS];
	char	 g_dontovw[NPARAMS];	/* don't overwrite */
	char	*g_outheaders[32];
	int	 g_outheaderN;
struct gateway	*g_parent;		/* parent class ... */

	char	*g_sender;
	char	*g_recipient;
	char	*g_rname; /* [xxx%] LOCAL [@domain] part of g_recipient */

	int	 g_option;
	char	 g_gwt[32];
	int	 g_count;

	FILE	*g_ts;
	FILE	*g_fs;
	char	*g_filter; /* saved global filter environment */
	char	 g_replyto[LNSIZE];
	char	 g_subject[LNSIZE];
} Gateway;

#define gv(px)		Gw->g_pv[px]
#define getp(px)	(gv(px) ? gv(px) : NULL)
#define getv(px)	((gv(px) && *gv(px)) ? gv(px) : NULL)

#define Option		Gw->g_option
#define Gwt		Gw->g_gwt
#define Sender		Gw->g_sender
#define Recipient	Gw->g_recipient
#define Rname		Gw->g_rname
#define Seqno		Gw->g_count
#define TS		Gw->g_ts
#define FS		Gw->g_fs
#define GFILTER		Gw->g_filter
#define ReplyTo		Gw->g_replyto
#define Subject		Gw->g_subject


static char *builtin_src[][8] = {
	/* MAIL TO NEWS */
	{
		"GATEWAY-TYPE: postnews",
		"SERVER-PROTO: nntp",
		"SERVER-HOST:  localhost",
		"Newsgroups:   junk",
		"Distribution: local",
	},
	/* MAIL TO MAIL */
	{
		"GATEWAY-TYPE: sendmail",
		"SERVER-PROTO: smtp",
		"SERVER-HOST:  localhost",
	},
	/* MAIL TO FTP */
	{
		"GATEWAY-TYPE: ftpmail",
		"SERVER-PROTO: ftp",
		"SERVER-HOST:  localhost",
	},
	/* MAIL TO HTTP */
	{
		"GATEWAY-TYPE: httpmail",
		"SERVER-PROTO: http",
		"SERVER-HOST:  localhost",
	},
	0
};
static Gateway builtins[8];


static Gateway *new_conf()
{
	return (Gateway*)calloc(1,sizeof(Gateway));
}
static Gateway *builtinGw(gwname,silent)
	char *gwname;
{	int gwi,pi,px;
	int nlen;
	char *name;

	if( gwname == NULL ){
		sv1log("#### NO GATEWAY TYPE SPECIFIED\n");
		return NULL;
	}
	if( builtins[0].g_pv[0] == NULL ){
		char **lines,*line,name[256],value[256],*vp;

		for( gwi = 0; (lines = builtin_src[gwi])[0]; gwi++ )
		for( pi = 0; line = lines[pi]; pi++ ){
			RFC822_decompField(line,name,value);
			vp = strip_spaces(value);
			px = param_idx(name);
			Verbose("SMTPGATE[%d][%d]=[%2d:%14s]=[%s]\n",
				gwi,pi,px,name,vp);
			if( px < 0 ){
				sv1log("#### UNKNOWN FILED IN BUILTIN SPEC: %s\n",
					line);
				return NULL;
			}
			builtins[gwi].g_pv[px] = strdup(vp);
		}
	}

	for( gwi = 0; name = builtins[gwi].g_pv[GWTYPE];  gwi++ ){
		nlen = strlen(name);
		if( strncasecmp(gwname,name,nlen) == 0 )
			if( gwname[nlen] == ':' || gwname[nlen] == 0 )
				return &builtins[gwi];
	}
	if( !silent )
		sv1log("#### unknown GATEWAY TYPE: %s\n",gwname);
	return NULL;
}
static copy_ifundef(Gw,Gws)
	Gateway *Gw,*Gws;
{	int pi;
	char *p1,*d1;

	for( pi = 1; pi < NPARAMS; pi++ )
		if( (p1 = getp(pi)) == NULL && (d1 = Gws->g_pv[pi]) )
			Gw->g_pv[pi] = d1;
}

static scan_conf(Gw,fp,log)
	FILE *fp;
	Gateway *Gw;
	FILE *log;
{	int size,rcc;
	char *buff;
	char *line,*next,lbuf[1024],*comment;
	char name[1024],*np,value[2048],*vp;
	int px;

	size = file_size(fileno(fp));
	buff = malloc(size+1);
	rcc = fread(buff,1,size,fp);
	buff[rcc] = 0;

	for( line = buff; line && *line; line = next ){
		next = nextField(line,1);
		if( (comment = strpbrk(line,"#\r\n")) && *comment == '#' ){
			if( comment == line )
				continue;
			else	*comment = 0;
		}

		RFC822_decompField(line,name,value);
		if( np = strchr(name,'/') )
			np = np + 1;
		else	np = name;
		vp = strip_spaces(value);

		if( 0 <= (px = param_idx(name)) ){
			if( px == OUT_HEADER ){
				lfprintf(log,NULL,"conf + %s/%s[%d]:%s\r\n",
					param_spec[px].p_pfix,param_spec[px].p_name,
					Gw->g_outheaderN,vp);
				Gw->g_outheaders[Gw->g_outheaderN++] = strdup(vp);
			}else
			if( Gw->g_pv[px] == 0 || px == INHERIT ){
				lfprintf(log,NULL,"conf + %s/%s:%s\r\n",
					param_spec[px].p_pfix,param_spec[px].p_name,vp);
				Gw->g_pv[px] = strdup(vp);
			}else{
				lfprintf(log,NULL,"conf - %s:%s\r\n",name,vp);
			}
		}else{
			lfprintf(log,NULL,"unknown -- %s:%s\r\n",name,vp);
		}
	}

	free(buff);
}

static char *user_class[8] = {
	"admin",
	"users",
	"",
	0
};

extern char *DELEGATE_SMTPGATE;
static char *SMTPGATE_DIR;

FILE *open_file(name,type,mode,rpath)
	char *name,*type,*mode,*rpath;
{	char cpath[1024];
	char *class;
	FILE *fp;
	int ui;

	if( SMTPGATE_DIR == NULL ){
		sv1log("No SMTPGATE directory specified.\n");
		return NULL;
	}

	for( ui = 0; class = user_class[ui]; ui++ ){
		sprintf(rpath,"%s/%s/%s",class,name,type);
		sprintf(cpath,"%s/%s",SMTPGATE_DIR,rpath);
		if( fp = fopen(cpath,mode) )
			return fp;
	}
	return NULL;
}

static load_conf(Conn,Gw,name,afp,sender,recipient,log)
	Connection *Conn;
	Gateway *Gw;
	char *name;
	FILE *afp;
	char *sender,*recipient;
	FILE *log;
{	char rpath[1024];
	FILE *cfp;

	if( SMTPGATE_DIR == NULL ){
		lfprintf(log,NULL,"No SMTPGATE directory specified.\n");
		return 0;
	}

	cfp = open_file(name,"conf","r",rpath);
	if( cfp != NULL ){
		lfprintf(log,NULL,"load config. file -- SMTPGATE/%s\n",rpath);
		scan_conf(Gw,cfp,log);
		fclose(cfp);
		return 1;
	}else{
		lfprintf(log,NULL,"cannot read config. file -- SMTPGATE/%s\n",
			rpath);
		return 0;
	}
}

/*
 *	SMTPGATE=dir[:list-of-subdirs]
 */
extern int (*LOG_substfile)();
scan_SMTPGATE(Conn,conf)
	Connection *Conn;
	char *conf;
{	char pb[1024];
	char *pv[NPARAMS+1],*p0,*p1;
	int pc,pi;
	int gwi;
	char dir[1024];

	if( builtinGw(POSTNEWS,0) == NULL ){
		sv1tlog("ERROR IN BUILTIN SPECIFICATION\n");
		Finish(1);
	}

	strcpy(pb,conf);
	pc = stoV(pb,NPARAMS,pv,':');
	p0 = pv[0];
	p1 = pv[1];

	strcpy(dir,p0);
	(*LOG_substfile)(dir,"",NULL,NULL,NULL);
	if( !fileIsdir(dir) ){
		sv1tlog("Warning: non existent SMTPGATE=%s [%s]\n",conf,dir);
		return -1;
	}else{
		sv1tlog("SMTPGATE=%s [%s]\n",conf,dir);
		SMTPGATE_DIR = strdup(dir);
		return 0;
	}
}

/*
 * OPTION
 */
#define RXO	0x01	/* Remove X-Original headers */
#define AXO	0x02	/* Append X-Original headers */
#define RES	0x04	/* Reject Empty Subject */
#define RNI	0x08	/* Reject No message-ID message */
#define REB	0x10	/* Reject Empby Body */
#define GWT	0x20	/* GateWay Trace */
#define ISN	0x40	/* Increment Sequence Number */
#define INH	0x80	/* INHibit to accepting at this name */

#define CHK_ACC_TO	"tmn"	/* check To: name matching */
#define MSG_NUMBER	"msn"	/* message number by posted article number */

static scan_options(opts,gwt,log)
	char *opts,*gwt;
	FILE *log;
{	int opti;
	char optb[1024];
	char *ov[32],*o1;
	int oc,oi;
	char msg[1024];

	opti = 0;
	strcpy(optb,opts);
	if( strchr(optb,',') )
		oc = stoV(optb,32,ov,',');
	else	oc = stoV(optb,32,ov,'.');

	for( oi = 0; oi < oc; oi++ ){
		o1 = strip_spaces(ov[oi]);

		if( strcaseeq(o1,"res") ){
			sprintf(msg,"Reject-Empty-Subject");
			opti |= RES;
		}else
		if( strcaseeq(o1,"reb") ){
			sprintf(msg,"Reject-Empty-Body");
			opti |= REB;
		}else
		if( strcaseeq(o1,"rxo") ){
			sprintf(msg,"Remove-X-Original");
			opti |= RXO;
		}else
		if( strcaseeq(o1,"axo") ){
			sprintf(msg,"Append-X-Original");
			opti |= AXO;
		}else
		if( strcaseeq(o1,"gwt") ){
			sprintf(msg,"Gateway-Trace");
			opti |= GWT;
			strcpy(gwt,o1);
		}else
		if( strcaseeq(o1,"isn") ){
			sprintf(msg,"Increment-Sequence-Number");
			opti |= ISN;
		}else{
			sprintf(msg,"unknown, ignored.",o1);
		}
		lfprintf(log,NULL,"OPTION: %s -- %s ON\r\n",o1,msg);
	}
	return opti;
}

static append_log(Conn,Gw,ts,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *ts,*log;
{	char proto[32];
	char *what;

	if( (Option & GWT) == 0 )
		return;

	fflush(log);
	ftruncate(fileno(log),ftell(log));
	fseek(log,0,0);

	strtoupper(DST_PROTO,proto);
	if( strcasecmp(proto,"smtp") != 0 )
		what = "gateway";
	else	what = "relay";

fprintf(ts,"\r\n");
fprintf(ts,"--MAILGATE %s/SMTP %s DeleGate/%s--\r\n",proto,what,DELEGATE_ver());
fprintf(ts,"Comment: This part including the CRLF at two lines above is a trace\r\n");
fprintf(ts,"  information appended by DeleGate which relayed this messaged from\r\n");
fprintf(ts,"  SMTP to %s.  The tracing is activated\r\n",proto);
 if( Gwt[0] ){
 fprintf(ts,"  by set \"gateway-trace\" option (%s) specified in the\r\n",Gwt);
 fprintf(ts,"  gateway for recipient address <%s>.\n",Recipient);
 }else{
 fprintf(ts,"  by a\r\n");
 fprintf(ts,"   X-Gateway-Control: trace-on\r\n");
 fprintf(ts,"  field included in the original message.\r\n");
 }
fprintf(ts,"\r\n");
copyfile1(log,ts);
fprintf(ts,"--\r\n");
}


subject_headclean(subj)
	char *subj;
{	char sc,*sp,*dp,tmp[2048];
	int withRe,numBracket;

	withRe = 0;
	numBracket = 0;

	for( sp = subj; sc = *sp; ){
		if( sc == ' ' || sc == '\t' ){
			sp++;
		}else
		if( sc == '[' && numBracket == 0 ){
			numBracket = 1;
			for( sp++; sc = *sp; sp++ )
				if( sc == ']' ){
					sp++;
					break;
				}
		}else
		if( strncasecmp(sp,"Re:",3) == 0 ){
			withRe++;
			sp += 3;
		}else
		if( strncasecmp(sp,"Re^",3) == 0 ){
			withRe++;
			for( sp += 3; *sp && isdigit(*sp); sp++ )
				;
			if( *sp == ':' )
				sp++;
		}else{
			break;
		}
	}
EXIT:
	strcpy(tmp,sp);
	subj[0] = 0;
	if( withRe )
		strcat(subj,"Re: ");
	strcat(subj,tmp);
}

static substitute(Gw,mfp,src,dst,log,sender)
	Gateway *Gw;
	char *src,*dst;
	FILE *log;
	char *sender;
{	char tmp[1024],val[1024];
	char *bp,*ep,*sp,*dp;

	strcpy(dst,src);
	strcpy(tmp,dst);

	if( strstr(dst,V_ERRORSTAT) ){
		strsubst(dst,V_ERRORSTAT,"something wrong...");
		lfprintf(log,NULL,"substitute -- %s => %s\n",tmp,dst);
	}
	if( strstr(dst,V_RECNAME) ){
		strsubst(dst,V_RECNAME,Rname);
		lfprintf(log,NULL,"substitute -- %s => %s\n",tmp,dst);
	}
	if( strstr(dst,V_SENDER) ){
		strsubst(dst,V_SENDER,sender);
		lfprintf(log,NULL,"substitute -- %s => %s\n",tmp,dst);
	}
	if( strstr(dst,V_SEQNO) ){
		sprintf(val,"%d",Seqno);
		strsubst(dst,V_SEQNO,val);
		lfprintf(log,NULL,"substitute -- %s => %s\n",tmp,dst);
	}
	if( strstr(dst,V_SUBJECT) ){
		fgetsHeaderField(mfp,"Subject",val,sizeof(val));
		strsubst(dst,V_SUBJECT,val);
		lfprintf(log,NULL,"substitute -- %s => %s\n",tmp,dst);
	}
	if( strstr(dst,V_CLEANSUBJ) ){
		fgetsHeaderField(mfp,"Subject",val,sizeof(val));
		subject_headclean(val);
		strsubst(dst,V_CLEANSUBJ,val);
		lfprintf(log,NULL,"substitute -- %s => %s\n",tmp,dst);
	}
}

static rewrite_header(Conn,Gw,mfp,ts,sender,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *mfp;
	FILE *ts;
	char *sender;
	FILE *log;
{	char *p1;
	int hi;
	char line[LNSIZE];

	if( p1 = getv(OUT_REPLY_TO) ){
		substitute(Gw,mfp,p1,ReplyTo,log,sender);
		lfprintf(log,ts,"Reply-To: %s\r\n",ReplyTo);
	}
	if( p1 = getv(OUT_SUBJECT) ){
		substitute(Gw,mfp,p1,Subject,log,sender);
		lfprintf(log,ts,"Subject: %s\r\n",Subject);
	}
	if( 0 < Gw->g_outheaderN ){
		for( hi = 0; hi < Gw->g_outheaderN; hi++ ){
			substitute(Gw,mfp,Gw->g_outheaders[hi],line,log,sender);
			lfprintf(log,ts,"%s\r\n",line);
		}
	}
}

/*
 *   MAIL_TO_NEWS
 *	"tc" is a stream connected to the SMTP client
 *	"mfp" is a file containing the data sent by DATA command.
 *	"ts" is a stream connected to the NNTP server
 */
static mail_to_news(Conn,Gw,tc,mfp,sender,recipient,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *tc,*mfp;
	char *sender,*recipient;
	FILE *log;
{	char line[1024],tmp[1024];
	char fname[1024],fvalue[1024];
	char myhost[256],clifhost[256],clhost[256],*dp;
	char path0[128];
	char To[1024],*to;
	char From[1024];
	char Newsgroups[1024],Distribution[1024];
	char EOH[8];
	char *p1;

	Newsgroups[0] = 0;
	Distribution[0] = 0;

	To[0] = 0;
	From[0] = 0;
	EOH[0] = 0;

	gethostname(myhost,sizeof(myhost));
	getFQDN(myhost,myhost);
	ClientIF_name(Conn,FromC,clifhost);
	getClientHostPort(Conn,clhost);

	if( (dp = strchr(sender,'@')) && dp[1] ){
		strcpy(path0,dp+1);
		strcat(path0,"!");
		sscanf(sender,"%[^@]",&path0[strlen(path0)]);
	}else	strcpy(path0,sender);
	lfprintf(log,TS,"Path: %s!SMTP-DeleGate%s!%s!%s!%s\r\n",
		myhost,DELEGATE_ver(),clifhost,clhost,path0);

	lfprintf(log,TS,
		"X-Forwarded: by %s (NNTP/SMTP gateway DeleGate/%s)\r\n",
		myhost,DELEGATE_ver());

	if( p1 = getv(OUT_NEWSGROUPS) ){
		substitute(Gw,mfp,p1,Newsgroups,log,sender);
		lfprintf(log,TS,"Newsgroups: %s\r\n",Newsgroups);
	}
	if( p1 = getv(OUT_DISTRIBUTION) ){
		substitute(Gw,mfp,p1,Distribution,log,sender);
		lfprintf(log,TS,"Distribution: %s\r\n",Distribution);
	}

	rewrite_header(Conn,Gw,mfp,TS,sender,log);

	/*
	 * should do MIME conversion ...
	 */
	for(;;){
		if( RFC822_fgetsHeaderField(line,sizeof(line),mfp) == NULL )
			break;

		if( line[0] == '\r' || line[0] == '\n' ){
			strcpy(EOH,line);
			break;
		}

		sscanf(line,"%[^: ]",fname);

		if( strcmp(fname,"To")   == 0 && (p1 = getv(ACC_TO))
		||  strcmp(fname,"From") == 0 && (p1 = getv(ACC_FROM))
		){
			getFieldValue(line,fname,fvalue);
			RFC822_strip_comment(fvalue,fvalue);
			if( tobe_rejected(tc,fname,fvalue,p1,log) != 0 )
				return -1;
		}

		if( strcasecmp(fname,"X-Gateway-Control") == 0 ){
			lfprintf(log,NULL,"control -- %s",line);
			if( strcasestr(line,"trace-on") )
				Option |=  GWT;
			else	Option &= ~GWT;
		}

		if( strcasecmp(fname,"Date") == 0 ){
			char oline[128];
			strcpy(oline,line);
			RFC822_strip_comment(oline,line);
			if( strcmp(line,oline) != 0 )
			lfprintf(log,NULL,"time zone erased -- %s",line);
		}

		if( strcasecmp(fname,"From") == 0 )
			getFieldValue(line,"From",From);

		if( strncasecmp(fname,"X-Original-",11) == 0 ){
			if( (Option&RXO) ){
				lfprintf(log,NULL,"removed -- %s",line);
				continue;
			}
		}

		if( strcaseeq(fname,"Path")
		 || Newsgroups[0]   && strcaseeq(fname,"Newsgroups")
		 || Distribution[0] && strcaseeq(fname,"Distribution")
		 || ReplyTo[0]      && strcaseeq(fname,"Reply-To")
		 || Subject[0]      && strcaseeq(fname,"Subject")
		 || strcaseeq(fname,"NNTP-Posting-Host")
		 || strcaseeq(fname,"Errors-To")
		 || strcaseeq(fname,"Return-Receipt-To")
		 || strcaseeq(fname,"Return-Path")
		 || strcaseeq(fname,"Received")
		 || strcaseeq(fname,"Apparently-To")
		 || strcaseeq(fname,"To")
		 || strcaseeq(fname,"Cc")
		){
			if( !(Option&AXO) ){
				lfprintf(log,NULL,"removed -- %s",line);
				continue;
			}

			lfprintf(log,NULL,"escaped -- X-Original-%s",line);
			fputs("X-Original-",TS);
		}
		if( strcaseeq(fname,"Sender") ){
			lfprintf(log,NULL,"escaped -- X-Original-%s",line);
			fputs("X-Original-",TS);
		}

		fputs(line,TS);

		if( strcasecmp(fname,"Subject") == 0 && Subject[0] == 0 )
			strcpy(Subject,line);
	}
	if( EOH[0] ){
		char *admin;

		if( From[0] && strcmp(From,sender) != 0 )
			lfprintf(log,TS,"Sender: %s\r\n",sender);

		if( admin = DELEGATE_MANAGER )
			lfprintf(log,TS,"Resent-Sender: %s\r\n",admin);

		if( Subject[0] == 0 && !(Option&RES) )
			lfprintf(log,TS,"Subject: (no subject)\r\n");

		fputs(EOH,TS);
	}

	for(;;){
		if( fgets(line,sizeof(line),mfp) == NULL )
			break;
		fputs(line,TS);
	}

	append_log(Conn,Gw,TS,log);
	return 0;
}

static nntp_post(Conn,Gw,tc,mfp,sender,recipient,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *tc,*mfp;
	char *sender;
	FILE *log;
{	char stat[1024];
	char msgid[1024],*dp;
	int off;

	if( fgetsHeaderField(mfp,"Message-Id",msgid,sizeof(msgid)) ){
		if( (dp = strchr(msgid,'@')) == NULL || dp[1] == 0 ){
			lfprintf(log,tc,"554 illegal Messae-ID: %s\r\n",msgid);
			return -1;
		}

		lfprintf(log,TS,"STAT %s\r\n",msgid);
		fflush(TS);

		fgets(stat,sizeof(stat),FS);
		lfprintf(log,NULL,"<<< %s",stat);
		if( atoi(stat) == 223 ){
			lfprintf(log,NULL,"duplicate Mesage-ID: %s",stat);
			return -1;
		}
		lfprintf(log,NULL,"ok - Mesage-ID: %s\r\n",msgid);
	}else{
		lfprintf(log,NULL,"no original Message-Id\r\n");
		/* if( Option & RNI ) ... REJECT */
	}

	lfprintf(log,TS,"POST\r\n");
	fflush(TS);

	if( fgets(stat,sizeof(stat),FS) == NULL ){
		lfprintf(log,tc,"421 NNTP server closed before POST.\r\n");
		return -1;
	}else{
		lfprintf(log,NULL,"<<< %s",stat);
	}
	if( atoi(stat) != 340 ){
		fprintf(tc,"554 posting failed: %s",stat);
		fflush(tc);
		return -1;
	}

	if( mail_to_news(Conn,Gw,tc,mfp,sender,recipient,log) < 0 )
		return -1;

	lfprintf(log,TS,".\r\n");
	fflush(TS);

	if( fgets(stat,sizeof(stat),FS) == NULL ){
		lfprintf(log,tc,"421 NNTP server closed during POST.\r\n");
		return -1;
	}

	lfprintf(log,NULL,"<<< %s",stat);
	if( atoi(stat) != 240 ){
		fprintf(tc,"554 posting failed: %s",stat);
		fflush(tc);
		return -1;
	}

	return 0;
}
static nntp_open(Conn,Gw,tc,mfp,recipient,pc,pv,log) 
	Connection *Conn;
	Gateway *Gw;
	FILE *tc;
	FILE *mfp;
	char *recipient;
	char *pv[];
	FILE *log;
{	int pi;
	char *p1;
	char *host;
	int port;
	char stat[1024];
	int server;

	for( pi = 0; pi < pc; pi++ ){
		char *p1,tag[128],val[128];
		int px;

		p1 = pv[pi];
		sscanf(p1,"%[^.].%s",tag,val);
		px = param_idx(tag);
		if( px < 0 ){
			lfprintf(log,tc,"config. error -- unknown [%s]\n",p1);
			return -1;
		}
		Gw->g_pv[px] = strdup(val);
	}

	host = getv(SERVHOST);
	if( p1 = getv(SERVPORT) )
		port = atoi(p1);
	else	port = 0;
	if( port == 0 )
		port = 119;

	lfprintf(log,NULL,"connecting to -- nntp://%s:%d\r\n",host,port);
	set_realserver(Conn,"nntp",host,port);

	if( setFilter(Conn,Gw,tc,mfp,log) != 0 )
		return -1;
	server = connect_to_serv(Conn,FromC,ToC,0);
	setFTOSV(GFILTER);
	if( server < 0 ){
		lfprintf(log,tc,"421 %s... cannot connect NNTP\r\n",recipient);
		fflush(tc);
		return -1;
	}

	FS = fdopen(FromS,"r");
	TS = fdopen(ToS,"w");

	fgets(stat,sizeof(stat),FS);
	lfprintf(log,NULL,"<<< %s",stat);

	return 0;
}
static postnews(Conn,Gw,tc,mfp,sender,recipient,pc,pv,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *tc,*mfp;
	char *sender;
	char *pv[];
	FILE *log;
{	int size;

	/* must skip header and see the size of body ... */
	size = file_size(fileno(mfp)) - ftell(mfp);
	lfprintf(log,NULL,"#### BODY SIZE = %d\n",size);
	if( size == 0 && Option & REB ){
		/* ... */
	}

	if( nntp_open(Conn,Gw,tc,mfp,recipient,pc,pv,log) != 0 )
		return -1;
	else
	if( nntp_post(Conn,Gw,tc,mfp,sender,recipient,log) != 0 )
		return -1;

	sendBCC(Conn,Gw,mfp,log);
	return 0;
}


static ftpmail(Conn,Gw,tc,mfp,sender,recipient,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *tc;
	FILE *mfp;
	char *recipient;
	FILE *log;
{	FILE *fpv[2],*ts,*fs;
	char host[256];
	int port;
	int rsock;
	char resp[1024];

	rsock = -1;
	port = 25;

	if( !localsocket(Conn->cl_sock) ){
		getClientHostPort(Conn,host); /* should use HELO ? */
		rsock = SMTP_open(Conn,fpv,host,port,sender,recipient,1,log);
	}
	if( rsock < 0 ){
		gethostname(host,sizeof(host));
		rsock = SMTP_open(Conn,fpv,host,port,sender,recipient,1,log);
	}
	if( rsock < 0 ){
		lfprintf(log,tc,"421 cannot open SMTP for response.\r\n");
		return -1;
	}

	fs = fpv[0];
	ts = fpv[1];
	fprintf(ts,"Subject: TEST RESPONSE\r\n");
	fprintf(ts,"\r\n");
	fprintf(ts,"(TEST RESPONSE)\r\n");
	lfprintf(log,ts,".\r\n");
	fflush(ts);
	SMTP_relay_stat(fs,NULL,resp);

	lfprintf(log,ts,"QUIT\r\n");
	fflush(ts);
	SMTP_relay_stat(fs,NULL,resp);
	return 0;
}

static put_maildata(Conn,Gw,ts,fs,mfp,resp,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *ts,*fs,*mfp;
	char *resp;
{	char *p1;
	char line[LNSIZE];
	char fname[LNSIZE];

	SMTP_putserv(log,fs,ts,resp,"DATA\r\n");
	if( atoi(resp) != 354 )
		return -1;

	rewrite_header(Conn,Gw,mfp,ts,Sender,log);

	for(;;){
		if( RFC822_fgetsHeaderField(line,sizeof(line),mfp) == NULL )
			break;

		sscanf(line,"%[^: ]",fname);
		if( strcaseeq(fname,"Subject" ) && Subject[0]
		 || strcaseeq(fname,"Reply-To") && ReplyTo[0]
		){
			fprintf(ts,"X-Original-");
		}
		fputs(line,ts);
		if( line[0] == '\r' || line[0] == '\n' )
			break;
	}
	copyfile1(mfp,ts);
	
	append_log(Conn,Gw,ts,log);
	lfprintf(log,ts,".\r\n");
	fflush(ts);
	SMTP_relay_stat(fs,NULL,resp);
	return atoi(resp);
}

FILE *TMPFILE();
extern FILE *expand_aliases();

sendmail_at(Conn,Gw,tc,mfp,host,port,recipients,sender,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *tc,*mfp;
	char *host,*recipients,*sender;
	FILE *log;
{	FILE *RCPT;
	char recipient[1024],*np;
	FILE *fpv[2],*ts,*fs;
	int rsock;
	char resp[1024];
	int nrec,nrec1,nerr,ri;
	int rcode = 0;

	RCPT = expand_aliases(recipients,log);
	if( RCPT == NULL ){
		lfprintf(log,tc,"421 error in alias expansion\r\n");
		return -1;
	}
	fgets(recipient,sizeof(recipient),RCPT);
	if( np = strpbrk(recipient,"\r\n") )
		*np = 0;
	rsock = SMTP_open(Conn,fpv,host,port,recipient,sender,0,log);
	setFTOSV(GFILTER);
	if( rsock < 0 ){
		lfprintf(log,tc,"421- cannot open SMTP for forwarding.\r\n");
		lfprintf(log,tc,"421  server = %s:%d.\r\n",host,port);
		return -1;
	}

	fs = fpv[0];
	ts = fpv[1];
	nrec = 1;
	nerr = 0;

	while( !feof(RCPT) ){
		for( nrec1 = 0; nrec1 < 64; nrec1++ ){
			if( fgets(recipient,sizeof(recipient),RCPT) == NULL )
				break;
			if( np = strpbrk(recipient,"\r\n") )
				*np = 0;
			lfprintf(log,ts,"RCPT To: %s\r\n",recipient);
			nrec++;
		}
		fflush(ts);
		for( ri = 0; ri < nrec1; ri++ ){
			SMTP_relay_stat(fs,NULL,resp);
			if( atoi(resp) != 250 ){
				lfprintf(log,NULL,"[%d] %s",ri,resp);
				nerr++;
			}
		}
	}
	if( nerr != 0 ){
		lfprintf(log,tc,"421 NOT sent: RCPT errors %d/%d\r\n",nerr,nrec);
		rcode = -1;
	}else{
		put_maildata(Conn,Gw,ts,fs,mfp,resp,log);
		lfprintf(log,NULL,"Sent to %d recipients\r\n",nrec);
	}

	lfprintf(log,ts,"QUIT\r\n");
	fflush(ts);
	SMTP_relay_stat(fs,NULL,resp);
	return rcode;
}
static sendmail(Conn,Gw,tc,mfp,sender,recipient,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *tc;
	FILE *mfp;
	char *sender,*recipient;
	FILE *log;
{	char *p1;
	char *host;
	int port;
	char *to,tobuf[1024];
	int off;

	to = getv(OUT_TO);
	if( to == NULL ){
		lfprintf(log,tc,"554 no forwarding info.\r\n");
		fflush(tc);
		return -1;
	}
	substitute(Gw,mfp,to,tobuf,log,sender);
	to = tobuf;

	host = getv(SERVHOST);
	if( p1 = getv(SERVPORT) ) 
		port = atoi(p1);
	else	port = 0;
	if( port == 0 )
		port = 25;

	if( setFilter(Conn,Gw,tc,mfp,log) != 0 )
		return -1;

	off = ftell(mfp);
	if( sendmail_at(Conn,Gw,tc,mfp,host,port,to,sender,log) < 0 )
		return -1;
	fseek(mfp,off,0);
	sendBCC(Conn,Gw,mfp,log);
	return 0;
}

/*
 *	BCC: <user@host.domain:port>
 */
static sendBCC(Conn,Gw,mfp,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *mfp;
	FILE *log;
{	char *bcc;
	char host[256],*sender;
	int port;

	bcc = getv(BCC);
	if( bcc == NULL )
		return;

	host[0] = 0;
	port = 0;
	sscanf(bcc,"%*[^@]@%[^:]:%d",host,&port);
	if( host[0] == 0 )
		strcpy(host,"localhost");
	if( port == 0 )
		port = 25;
	sender = DELEGATE_MANAGER;

	lfprintf(log,NULL,"sending BCC -- %s [smtp://%s:%d] sender=%s\n",
		bcc,host,port,sender);

	sendmail_at(Conn,Gw,log,mfp,host,port,bcc,sender,log);
}

static FILE *lock_counter(name,mfp,countp,log)
	char *name;
	FILE *mfp;
	int *countp;
	FILE* log;
{	char rpath[1024];
	FILE *counter;
	int opid,otime;
	char line[1024];
	int count;

	*countp = 0;
	if( *name == 0 )
		return NULL;

	counter = open_file(name,"count","r+",rpath);
	if( counter != NULL ){
		count = 0;
		opid = 0;
		fgets(line,sizeof(line),counter);
		sscanf(line,"%d pid=%d time=%d",&count,&opid,&otime);
		if( count == 0 )
			count = 1;
	}else
	if( counter = open_file(name,"count","w+",rpath) ){
		lfprintf(log,NULL,"created counter -- SMTPGATE/%s\r\n",rpath);
		fprintf(counter,"1 pid=%d\r\n",getpid());
		fflush(counter);
		count = 1;
		opid = getpid();
		otime = time(0);
	}else{
		sv1log("#### cannot open COUNTER: %s\r\n",rpath);
		return NULL;
	}

	if( lock_exclusiveTO(fileno(counter),60*1000,NULL) ){
		sv1log("#### cannot lock COUNTER: %s\r\n",rpath);
		return NULL;
	}

	lfprintf(log,NULL,"got Counter -- %d : SMTPGATE/%s\n",count,rpath);
	*countp = count;
	return counter;
}
static unlock_counter(counter,count,inc,log)
	FILE *counter;
	FILE *log;
{
	if( inc ){
		fseek(counter,0,0);
		count += inc;
		fprintf(counter,"%d pid=%d time=%d\r\n",
			count,getpid(),time(0));
		lfprintf(log,NULL,"set Counter -- %d\r\n",count);
	}
	fclose(counter);
}

static lfprintf1(log,fmt,a,b,c,d,e,f,g)
	char *fmt,*a,*b,*c,*d,*e,*f,*g;
{
	lfprintf(log,NULL,fmt,a,b,c,d,e,f,g);
}

static match_aclfile(what,addr,acladdr,afp,log)
	char *what,*addr,*acladdr;
	FILE *afp;
	FILE *log;
{	char line[1024],*vp,*dp;
	int nlines;

	for( nlines = 1; fgets(line,sizeof(line),afp) != NULL; nlines++ ){
		if( strcasestr(line,addr) == NULL )
			continue;

		sv1log("#### found -- <%s> at line#%d of %s\n",addr,nlines,acladdr);
		if( dp = strpbrk(line,"\r\n") )
			*dp = 0;
		lfprintf(log,NULL,"accept %s -- <%s> is in the ACL as '%s'\r\n",
			what,addr,line);
		return 1;
	}
	sv1log("#### not found -- <%s> in %s\n",addr,acladdr);
	return 0;
}
static check_infile(tc,what,addr,acl,log)
	FILE *tc;
	char *what,*addr,*acl;
	FILE *log;
{	FILE *afp;

	afp = fopen(acl,"r");
	if( afp == NULL ){
		lfprintf(log,tc,"554 cannot open the ACL file.\n");
		fflush(tc);
		return -1;
	}
	return match_aclfile(what,addr,acl,afp,log);
}
static check_byexpn(tc,what,addr,acladdr,log)
	FILE *tc;
	char *what,*addr,*acladdr;
	FILE *log;
{	FILE *afp;

	afp = SMTP_getEXPN(acladdr);
	if( afp == NULL ){
		lfprintf(log,tc,"554 cannot get EXPN <%s>\n",addr);
		fflush(tc);
		return -1;
	}
	return match_aclfile(what,addr,acladdr,afp,log);
}
static tobe_rejected(tc,what,addr,acli,log)
	FILE *tc;
	char *what,*addr,*acli;
	FILE *log;
{	char *av[256],*acla,*aclb,*aclc,*aclp,*acl0,*acl1;
	char acladdr[256];
	int ac,ai,negate,match,subst;

	acla = strdup(acli);
	aclb = strdup(acli);
	aclc = malloc(strlen(acli)+1);
	aclp = aclc;
	ac = stoV(acla,256,av,',');

	for( ai = 0; ai < ac; ai++ ){
		acl1 = av[ai];
		acl1 = acl0 = strip_spaces(acl1);
		if( negate = (*acl1 == '!') )
			acl1++;

		if( subst = isFullpath(acl1) ){
			match = check_infile(tc,what,addr,acl1,log);
			if( match < 0 )
				goto EXIT;
		}else
		if( subst = (strncmp(acl1,"${expn:",7)==0) ){
			acladdr[0] = 0;
			sscanf(acl1,"${expn:%[^}]}",acladdr);
			match = check_byexpn(tc,what,addr,acladdr,log);
			if( match < 0 )
				goto EXIT;
		}
		if( subst ){
			if( match ){
				if( aclp != aclc) *aclp++ = ',';
				if( negate )
					*aclp++ = '!';
				strcpy(aclp,"*");
			}
		}else{
			if( aclp != aclc) *aclp++ = ',';
			strcpy(aclp,acl0);
		}
		aclp += strlen(aclp);
	}
	lfprintf(log,NULL,"preprocessed ACL: [%s] -> [%s]\n",acli,aclc);

	if( match = strmatch_list(addr,aclc,"",lfprintf1,log) ){
		lfprintf(log,NULL,"accept %s -- %s is in [%s]\r\n",
			what,addr,aclc);
	}else{
		lfprintf(log,NULL,"reject %s -- %s is not in [%s]\r\n",
			what,addr,aclc);
		lfprintf(log,tc,"554 Forbidden by rule\r\n");
		fflush(tc);
	}
EXIT:
	free(acla);
	free(aclb);
	free(aclc);
	return match ? 0 : -1;
}

extern char *getFTOSV();
static setFilter(Conn,Gw,tc,mfp,log)
	Connection *Conn;
	Gateway *Gw;
	FILE *tc;
	FILE *mfp;
	FILE *log;
{	char *filter;

	filter = getv(OUT_FILTER);
	if( filter == NULL )
		return 0;

	lfprintf(log,NULL,"FILTER: %s\r\n",filter);
	scan_FTOSV(filter);
	return 0;
}

SMTPgateway(Conn,tc,mfp,sender,recipient,log)
	Connection *Conn;
	FILE *tc;
	FILE *mfp;
	char *sender,*recipient;
	FILE *log;
{	Gateway *Gw,*BGw;
	char local[1024],domain[1024];
	char *pv[128],pb[1024],*p1,*gwname,*parent;
	int pc,pi;
	char params[1024],*local1,*local2,*pp;
	int lev;
	int rcode = 0;
	FILE *counter,*log1;

	/*
	 *  search and select configuration file baed on
	 *  "sender", "recipients" and "mfp"
	 */
	local[0] = domain[0] = 0;
	sscanf(recipient,"%[^@]@%s",local,domain);
	strcpy(pb,local);
	pc = stoV(pb,128,pv,'%');
	gwname = local1 = local2 = pv[--pc];
	if( gwname == NULL )
		return 0;

	pp = params;
	for( pi = 0; pi < pc; pi++ ){
		sprintf(pp,"[%s]",pv[pi]);
		pp += strlen(pp);
	}
	lfprintf(log,NULL,"GWNAME = %s %s\n",gwname,params);

	Gw = new_conf();
	Sender = strdup(sender);
	Recipient = strdup(recipient);
	Rname = strdup(local1);

	if( load_conf(Conn,Gw,gwname,mfp,sender,recipient,log) == 0 ){
		lfprintf(log,NULL,"GWNAME = %s (default gateway)\n",DEFAULT);
		if( load_conf(Conn,Gw,DEFAULT,mfp,sender,recipient,log) ){
			if( parent = getv(INHERIT) )
				gwname = parent;
			lfprintf(log,NULL,"GWNAME = %s\n",gwname);
			local2 = DEFAULT;
		}
	}
	Gw->g_pv[GWTYPE] = strdup(gwname);

	/*
	 *  load parent if exists
	 */
	for( lev = 0; lev < 4; lev++ ){
		parent = getv(INHERIT);
		if( parent == NULL )
			break;

		if( BGw = builtinGw(parent,1) ){
			gwname = parent;
			break;
		}

		lfprintf(log,NULL,"GWNAME = %s\n",parent);
		if( load_conf(Conn,Gw,parent,mfp,sender,recipient,log) == 0 )
			break;
		parent = getv(INHERIT);
		if( parent != NULL )
			gwname = parent;
	}

	/*
	 *  builtin class
	 */
	if( (BGw = builtinGw(gwname,0)) == NULL ){
		lfprintf(log,NULL,"unknown Gateway Name -- %s\r\n",gwname);
		return 0;
	}
	Gw->g_parent = BGw;
	copy_ifundef(Gw,Gw->g_parent);
	lfprintf(log,NULL,"GWNAME = %s (built-in)\n",gwname);

	/*
	 *  access control
	 */
	if( p1 = getv(ACC_SENDER) ){
		if( tobe_rejected(tc,"Sender",sender,p1,log) != 0 ){
			rcode = -1;
			goto EXIT;
		}
	}
	if( p1 = getv(ACC_RECIPIENT) ){
		if( tobe_rejected(tc,"Recipient",recipient,p1,log) != 0 ){
			rcode = -1;
			goto EXIT;
		}
	}

	/*
	 *  save FTOSV filter
	 */
	GFILTER = getFTOSV();

	/*
	 *  call gateways
	 */
	if( p1 = getv(OPTION) )
		Option = scan_options(p1,Gwt,log);
	else	Option = 0;
	counter = NULL;

	if( Option & ISN ){
		counter = lock_counter(local2,mfp,&Seqno,log);
		if( counter == NULL ){
			lfprintf(log,tc,"421 cannot open/lock counter.\r\n");
			rcode = -1;
			goto EXIT;
		}
	}

	if( strcaseeq(gwname,POSTNEWS) ){
		if( postnews(Conn,Gw,tc,mfp,sender,recipient,pc,pv,log) == 0 )
			rcode =  1;
		else	rcode = -1;
	}else
	if( strcaseeq(gwname,SENDMAIL) ){
		if( sendmail(Conn,Gw,tc,mfp,sender,recipient,log) == 0 )
			rcode =  1;
		else	rcode = -1;
	}else
	if( strcaseeq(gwname,FTPMAIL) ){
		if( ftpmail(Conn,Gw,tc,mfp,sender,recipient,log) == 0 )
			rcode =  1;
		else	rcode = -1;
	}

	if( counter != NULL  )
		unlock_counter(counter,Seqno,rcode==1,log);

EXIT:
	lfprintf(log,NULL,"END(%d) %s <- %s\n",rcode,recipient,sender);
	if( rcode != 0 ){
		char rpath[1024];
		if( log1 = open_file(local2,"log","a",rpath) ){
			fflush(log);
			fseek(log,0,0);
			copyfile1(log,log1);
			close(log1);
			fseek(log,0,2);
			sv1log("#### wrote logfile of %s : %s\n",local1,rpath);
		}else{
			sv1log("#### cannot open logfile of %s : %s\n",local1,rpath);
		}
	}
	return rcode;
}
