/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1994 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	textconv.c (text / character code conversion)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:

    codeconv_spec = outcode [ "." incode ]

    outcode       = codename [ "_" extension ]

    incode        = codename [ "_" extension ]

    codename      = [0-9A-Za-z]+

    extension     = [0-9A-Za-z]+

    Example:

    	jis		== jis.jp : japanese input to  JIS (ISO-2022-JP)
	cii		== cii_gif.any : any input to CII
	jis.cn		== Chinese {BIG5,GB,HZ} to JIS by CN2JP
	euc.big5	== Chinese {BIG5,HZ} to EUC by CN2JP

History:
	941001 extracted from conf.c and gopher.c
	960105 introduced codeconv_spec and cascaded conversions

TODO:
	MIME (header) encoding to CII
//////////////////////////////////////////////////////////////////////#*/
#include "delegate.h"

int suppress_codeconv;
extern int PLAIN_TO_HTML_PRE;
extern int repair_JIS;
extern char *CII_ubase;
extern char *CII_srcurl;

#define CC_THRU	'='	/* Through, without conversion (output only) */
#define CC_ANY	'*'	/* any encoding (input only) */
#define CC_HTML	'H'	/* HTML character encoding */
#define CC_EUC	'E'	/* EUC (EUC-JP) */
#define CC_JIS	'J'	/* JIS (ISO-2022-JP) */
#define CC_SJIS	'S'	/* SJIS (Shift-JIS, MS-Kanji) */
#define CC_II	'I'	/* Inline Image in HTML (output only) */
#define CC_BIG5	'b'	/* Big5 */
#define CC_CN	'c'	/* Chinese {Big5,GB,HZ} (input only) */
#define CC_GB	'g'	/* GB */
#define CC_HZ	'h'	/* HZ */
#define CC_GRE	'G'	/* Greek */
#define CC_CYR	'Y'	/* Cyrillic (ISO-8859-5) */
#define CC_KOI8	'K'	/* Cyrillic (KOI8-R) */
#define CC_RECV	'R'	/* Recover lost ESCs */
#define CC_FIX	'F'	/* Fix lost ESC $ B */

#define JIS_BUFSIZ(sz)	(sz * 2 + 1000)

typedef struct {
	char	*cv_what; /* realm of the conversion */
	char	*cv_spec; /* conversion spec. source */
	int	 cv_occe; /* output charcode encoding */
	char	*cv_occs; /* output charcode spec. */
	int	 cv_icce; /* input charcode encoding */
	char	*cv_iccs; /* input charcode spec. */
} Conv;

static Conv CodeConv[64] = {{"*","THRU",CC_THRU,"=",CC_ANY,"*"}};
#define CONVX0	1
int CodeConv_X = CONVX0;	/* global */
int CodeConv_x;		/* local to connection */
#define CONVX	(CodeConv_X + CodeConv_x)

#define JP_CODE(ce) \
	(  strcaseeq(ce,"JP") || strcaseeq(ce,"JIS") \
	|| strcaseeq(ce,"SJIS") || strcaseeq(ce,"EUC") )

#define JP_code(ce) \
	( ce == CC_EUC || ce == CC_JIS || ce == CC_SJIS )

#define CN_CODE(ce) \
	(  strcaseeq(ce,"CN") || strcaseeq(ce,"HZ") \
	|| strcaseeq(ce,"GB") || strcaseeq(ce,"BIG5") ) \

#define CN_code(ce) \
	( ce == CC_CN || ce == CC_HZ || ce == CC_GB || ce == CC_BIG5 )

#define CCX_code(ce) \
	( CN_code(ce) || ce == CC_GRE || ce == CC_CYR || ce == CC_KOI8 )

static scan_CHARCODE1(charcode,local)
	char *charcode;
{	char ccb[256],*incode,*outcode,*dp;
	char spec[256];

	strcpy(ccb,charcode);
	if( dp = strchr(ccb,'/') ){
		*dp++ = 0;
		outcode = dp;
		incode = ccb;
	}else
	if( dp = strchr(ccb,'.') ){
		*dp++ = 0;
		incode = dp;
		outcode = ccb;
	}else{
		incode = "JP";
		outcode = ccb;
	}

	sprintf(spec,"%s.%s",outcode,incode);
	scan_CODECONV(spec,CCV_TOCL,local);
	return 0;
}
scan_CHARCODE(Conn,charcodes)
	Connection *Conn;
	char *charcodes;
{
	scan_commaList(charcodes,0,scan_CHARCODE1,0);
}

scan_codename(str,name)
	char *str,*name;
{
	anwordscan(str,name);
}
scan_CODECONV(spec,what,local)
	char *spec,*what;
{	Conv *cv;
	char *dp,outcx[128],outcs[128],incx[128],incs[128];
	char *cvspec;
	int oc,ic;

	if( 0 < CONVX && strcmp(CodeConv[CONVX-1].cv_spec,spec) == 0 ){
		sv1log("## ignored the same [%d] CODECONV=%s\n",CONVX,spec);
		return 0;
	}

	cv = &CodeConv[CONVX];

	sscanf(spec,"%[^.]",outcx);
	scan_codename(spec,outcs);

	strcpy(incs,"jp");
	strcpy(incx,"jp");
	if( dp = strchr(spec,'.') ){
		dp++;
		if( strncmp(dp,"cc=",3) == 0 )
			dp += 3;

		if( dp[0] ){
			sscanf(dp,"%[^.]",incx);
			if( strncmp(dp,"cc=",3) == 0 )
				dp += 3;
			scan_codename(dp,incs);
		}
	}

	cv->cv_what = what;
	cv->cv_spec = strdup(spec);
	cv->cv_occs = strdup(outcx);
	cv->cv_iccs = strdup(incx);

	if( strcaseeq(outcs, "html") )	oc = CC_HTML;	else
	if( strcaseeq(outcs, "fix" ) )  oc = CC_FIX;	else
	if( strcaseeq(outcs, "ii"  ) )	oc = CC_II;	else
	if( strcaseeq(outcs, "jp"  ) )  oc = CC_JIS;	else
	if( strcaseeq(outcs, "jis" ) )  oc = CC_JIS;	else
	if( strcaseeq(outcs, "euc" ) )  oc = CC_EUC;	else
	if( strcaseeq(outcs, "sjis") )  oc = CC_SJIS;	else
	if( strcaseeq(outcs, "cn"  ) )	oc = CC_HZ;	else
	if( strcaseeq(outcs, "hz"  ) )	oc = CC_HZ;	else
	if( strcaseeq(outcs, "gb"  ) )	oc = CC_GB;	else
	if( strcaseeq(outcs, "big5") )	oc = CC_BIG5;	else
	if( strcaseeq(outcs, "gre") )	oc = CC_GRE;	else
	if( strcaseeq(outcs, "cyr") )	oc = CC_CYR;	else
	if( strcaseeq(outcs, "koi8") )	oc = CC_KOI8;	else
					oc = CC_THRU;
	cv->cv_occe = oc;

	if( strcaseeq(incs, "html") )	ic = CC_HTML;	else
	if( strcaseeq(incs, "gre") )	ic = CC_GRE;	else
	if( strcaseeq(incs, "cyr") )	ic = CC_CYR;	else
	if( strcaseeq(incs, "koi8") )	ic = CC_KOI8;	else
	if( CN_CODE(incs) )		ic = CC_CN;	else
					ic = CC_JIS;
	cv->cv_icce = ic;

	sv1log("CODECONV[%d](%s,%s,%s) => %s.%s\n",CONVX,
		local?"local":"global",what,spec, outcx, incx);

	if( local )
		CodeConv_x++;
	else	CodeConv_X++;
	return 0;
}

static to_CII(src,dst,ctype)
	char *src,*dst,*ctype;
{	char *csv[32];
	int csn,cx;
	Conv *cv;

	csn = 0;
	for( cx = CONVX-1; 0 <= cx; cx-- ){
		cv = &CodeConv[cx];
		if( cv->cv_occe == CC_II )
			csv[csn++] = cv->cv_iccs;
	}
	csv[csn] = 0;
	if( csn )
		CII_toCII(src,dst,ctype,csv,CII_ubase,CII_srcurl);
	else	strcpy(dst,src);
}

static convenv(what,cvv)
	char *what;
	Conv *cvv[];
{	Conv *cv;
	int cvi,cvj,ncv;

	ncv = 0;

	/* for backward compatiblity */
	if( CONVX == CONVX0 ){
		int ic,oc;
		static Conv cv0;

		ic = oc = 0;
		if( strchr(DELEGATE_FLAGS,'c') ) ic = CC_CN;
		if( strchr(DELEGATE_FLAGS,'E') ) oc = CC_EUC;  else
		if( strchr(DELEGATE_FLAGS,'S') ) oc = CC_SJIS; else
		if( strchr(DELEGATE_FLAGS,'J') ) oc = CC_JIS;  else
		if( strchr(DELEGATE_FLAGS,'I') ) oc = CC_II;

		if( ic != 0 || oc != 0 ){
			cv0 = CodeConv[0];
			if( ic ){ cv0.cv_icce = ic; }
			if( oc ){ cv0.cv_occe = oc; }
			cvv[ncv++] = &cv0;
		}
	}

	for( cvi = CONVX-1; 0 <= cvi; cvi-- ){
		cv = &CodeConv[cvi];
		if( cv->cv_occe == CC_THRU && 0 < ncv )
			break;

		for( cvj = 0; cvj < ncv; cvj++ )
			if( cvv[cvj]->cv_icce == cv->cv_icce )
				goto NEXT;

		cvv[ncv++] = cv;
		if( cv->cv_occe == CC_THRU )
			break;
	NEXT:;
	}

	return ncv;
}

static char *ccesym(cce)
{
	switch( cce ){
		case CC_II:   return "II";
		case CC_FIX:  return "FIX";
		case CC_SJIS: return "SJIS";
		case CC_JIS:  return "JIS";
		case CC_EUC:  return "EUC";
		case CC_CN:   return "CN";
		case CC_HZ:   return "HZ";
		case CC_GB:   return "GB";
		case CC_BIG5: return "BIG5";
		default:      return "THRU";
	}
}

plain2html()
{	int ncv,cvi;
	Conv *cvv[16];

	if( PLAIN_TO_HTML_PRE )
		return 1;

	ncv = convenv(CCV_TOCL,cvv);
	for( cvi = 0; cvi < ncv; cvi++ )
		if( cvv[cvi]->cv_occe == CC_II )
			return 1;
	return 0;
}

static codeconvs(what,cvenv)
	char *what,*cvenv;
{	Conv *cvv[16],*cv;
	int ncv,cvi;
	char *sp;

	ncv = convenv(CCV_TOCL,cvv);
	if( cvenv ){
		sp = cvenv;
		for( cvi = ncv-1; 0 <= cvi; cvi-- ){
			cv = cvv[cvi];
			sprintf(sp,"%s.%s",cv->cv_occs,cv->cv_iccs);
			sp += strlen(sp);
			if( 0 < cvi ){
				sprintf(sp,",");
				sp += strlen(sp);
			}
		}
		*sp = 0;
	}
	return cvv[0]->cv_occe;
}

cur_codeconvCL(cvenv)
	char *cvenv;
{	int occe;

	occe = codeconvs(CCV_TOCL,cvenv);
	if( occe == 0 || occe == CC_THRU )
		return 0;
	else	return occe;
}

line_codeconv_repair(src,dst,ctype)
	char *src,*dst,*ctype;
{
	repair_JIS = 1;
	line_codeconv(src,dst,ctype);
	repair_JIS = 0;
}

codeconv_line(src,dst,ctype,recv)
	char *src,*dst,*ctype;
{
	if( recv )
		line_codeconv_repair(src,dst,ctype);
	else	line_codeconv(src,dst,ctype);
}

extern char *malloc();
line_codeconv(src,dst,ctype)
	char *src,*dst,*ctype;
{	Conv *cvv[16],*cv;
	int ncv,cvi;
	char *tmpv[2],*tmp;
	int tmpx,tx;

	if( suppress_codeconv ){
		strcpy(dst,src);
		return;
	}
	if( src != dst )
		dst[0] = 0;

	ncv = convenv(CCV_TOCL,cvv);
	for( cvi = 0; cvi < ncv; cvi++ ){
		cv = cvv[cvi];
		tmpx = 0;
		if( src == dst )
			tmpv[tmpx++] = src = strdup(src);

		if( CCX_code(cv->cv_icce) && JP_code(cv->cv_occe) ){
			tmpv[tmpx++] = tmp = malloc(JIS_BUFSIZ(strlen(src)));
			ccx(src,tmp,cv->cv_iccs,cv->cv_occs,ctype);
			src = tmp;
		}else
		if( JP_code(cv->cv_icce) && CCX_code(cv->cv_occe) ){
			/* JIS TO another code ... */
		}

		switch( cv->cv_occe ){
		case CC_FIX:  FIX_2022(src,dst,ctype);break;
		case CC_EUC:  TO_EUC( src,dst,ctype); break;
		case CC_SJIS: TO_SJIS(src,dst,ctype); break;
		case CC_JIS:  TO_JIS( src,dst,ctype); break;
		case CC_II:   to_CII( src,dst,ctype); break;
		default:      strcpy( dst,src);       break;
		}

	NEXT:	for( tx = 0; tx < tmpx; tx++ )
			free(tmpv[tx]);
		src = dst;
	}
}

do_codeconv(ccode,src,dst,ctype)
	char *ccode,*src,*dst,*ctype;
{
	if( ccode == CCV_TOCL )
		line_codeconv(src,dst,ctype);
	else
	if( ccode == CCV_TOSV )
		TO_JIS(src,dst,ctype);
	else
	switch( ccode[0]  ){
		case CC_FIX:  FIX_2022(src,dst,ctype);break;
		case CC_EUC:  TO_EUC(src,dst,ctype);  break;
		case CC_SJIS: TO_SJIS(src,dst,ctype); break;
		case CC_JIS:  TO_JIS(src,dst,ctype);  break;
		default:      strcpy(dst,src); break;
	}
}

char *codeconv_charset()
{
	switch( cur_codeconvCL(NULL) ){
		case CC_JIS:  return "iso-2022-jp";
		case CC_EUC:  return "x-euc-jp";
		case CC_SJIS: return "x-sjis";
	}
	return NULL;
}

codeconv_toCII(spec)
	char *spec;
{
	if( strncasecmp(spec,"II",2) == 0 )
		return 1;
	else	return 0;
}

codeconv_bufsize(ccode,size)
{
	switch( ccode ){
		case CC_FIX:  return JIS_BUFSIZ(size);
		case CC_JIS:  return JIS_BUFSIZ(size);
		case CC_II:   return size * 40;
		case CC_EUC:  return size;
		case CC_SJIS: return size;
	}
	return size;
}

check_codeconv(dolog){
	char cvenv[128];
	int tocl;

	if( tocl = cur_codeconvCL(cvenv) ){
		if( dolog ){
			sv1log("Code Conversion [CHARCODE=%s]\n",cvenv);
		}
		return tocl;
	}
	return 0;
}

codeconv_get(ctype,xcharset, p2h)
	char *ctype;
	char **xcharset;
	int *p2h;
{
	if( cur_codeconvCL(NULL) ){
		if( xcharset ) *xcharset = codeconv_charset();
		if( p2h      ) *p2h = plain2html();
		return 1;
	}else{
		if( xcharset ) *xcharset = 0;
		if( p2h      ) *p2h = 0;
		return 0;
	}
}
codeconv_set(enable,charcode,p2h)
	char *charcode;
{	int osupp;

	osupp = suppress_codeconv;
	if( enable != -1 )
		suppress_codeconv = !enable;
	if( charcode != NULL )
		scan_CHARCODE1(charcode,0);
	if( p2h != -1 )
		PLAIN_TO_HTML_PRE = p2h;
	return !osupp;
}


need_CIIfont(convspec,font)
	char *convspec,*font;
{	char *dp;
	char word1[128];

	scan_codename(convspec,word1);
	if( strcmp(word1,"II") == 0 ){
		strcpy(font,"JISX0208");
		return 1;
	}
	if( (dp = strstr(convspec,"/II")) && !Isalnum(dp[3]) ){
		strcpy(font,word1);
		return 1;
	}
	return 0;
}

need_CCXtable(convspec,charset)
	char *convspec,*charset;
{	char word1[128];
	char *dp,oce[128];

	scan_codename(convspec,word1);
	if( dp = strchr(convspec,'/') )
		scan_codename(dp+1,oce);
	else	strcpy(oce,word1);

	if( CN_CODE(word1) ){
		if( JP_CODE(oce) ){
			strcpy(charset,"CN-JP");
			return 1;
		}
	}
	return 0;
}

check_codeconvSP(Conn,dolog)
	Connection *Conn;
{
	if( BORN_SPECIALIST || ACT_SPECIALIST || ACT_TRANSLATOR )
		return check_codeconv(dolog);
	else	return 0;
}

extern char *fgetsTIMEOUT();
CCV_relay_texts(Conn,ins,out,dup)
	Connection *Conn;
	FILE *ins[],*out,*dup;
{	char *rs,line[4096];
	char xline[4096];
	int totalc;
	int do_conv;
	FILE *in;
	int inx,pending;

	do_conv = check_codeconvSP(Conn,1);
	totalc = 0;

	inx = 0;
	if( in = ins[inx] )
		inx++;
	pending = 0;

	while( in != NULL ){
		if( fPollIn(in,100) == 0 )
			fflushTIMEOUT(out);

		rs = fgetsTIMEOUT(line+pending,sizeof(line)-pending,in);
		if( rs == NULL ){
			if( in = ins[inx] ){
				inx++;
				continue;
			}
			if( pending == 0 )
				break;
		}else{
			if( line[pending] != 0 ) /* line buff. is not full */
			if( strpbrk(line,"\r\n") == 0 ){
				pending = strlen(line);
				continue;
			}
		}

		pending = 0;
		totalc += strlen(line);
		if( do_conv ){
			line_codeconv(line,xline,"*/*");
			fputs(xline,out);
		}else	fputs(line,out);

		if( dup )
			fputs(line,dup);
	}
	return  totalc;
}
CCV_relay_text(Conn,in,out,dup)
	Connection *Conn;
	FILE *in,*out,*dup;
{	FILE *ins[2];

	ins[0] = in;
	ins[1] = NULL;
	return CCV_relay_texts(Conn,ins,out,dup);
}
CCV_relay_textX(ccspec,in,out)
	char *ccspec;
	FILE *in,*out;
{	Connection ConnBuf, *Conn = &ConnBuf;

	bzero(Conn,sizeof(Connection));
	CodeConv_X = CONVX0;
	CodeConv_x = 0;
	scan_CHARCODE(Conn,ccspec);
	ACT_TRANSLATOR = 1;
	return CCV_relay_text(Conn,in,out,NULL);
}
